@@ -1,6 +1,6 @@
 /*
- * Copyright(c) 2012 Tim Ruehsen
- * Copyright(c) 2015-2019 Free Software Foundation, Inc.
+ * Copyright (c) 2012 Tim Ruehsen
+ * Copyright (c) 2015-2023 Free Software Foundation, Inc.
  *
  * This file is part of Wget.
  *
@@ -49,7 +49,6 @@ void job_free(JOB *job)
 	wget_vector_free(&job->parts);
 	wget_list_free(&job->remaining_sig_ext);
 	xfree(job->sig_req);
-	xfree(job->local_filename);
 	xfree(job->sig_filename);
 }
 
@@ -122,7 +121,7 @@ static int check_file_hash(HASH *hash, const char *fname)
 }
 */
 
-static int _check_file_fd(wget_metalink_hash *hash, int fd)
+static int check_file_fd(wget_metalink_hash *hash, int fd)
 {
 	char sum[128 + 1]; // large enough for sha-512 hex
 
@@ -196,7 +195,7 @@ int job_validate_file(JOB *job)
 		for (int it = 0; errno != EINTR && it < wget_vector_size(metalink->hashes); it++) {
 			wget_metalink_hash *hash = wget_vector_get(metalink->hashes, it);
 
-			if ((rc = _check_file_fd(hash, fd)) == -1)
+			if ((rc = check_file_fd(hash, fd)) == -1)
 				continue; // hash type not available, try next
 
 			break;
@@ -232,7 +231,7 @@ int job_validate_file(JOB *job)
 
 			part.id = it + 1;
 
-			if ((rc = check_piece_hash(hash, fd, part.position, part.length)) != 1) {
+			if (check_piece_hash(hash, fd, part.position, part.length) != 1) {
 				info_printf(_("Piece %d/%d not OK - requeuing\n"), it + 1, wget_vector_size(metalink->pieces));
 				wget_vector_add_memdup(job->parts, &part, sizeof(PART));
 				debug_printf("  need to download %llu bytes from pos=%llu\n",
@@ -277,7 +276,7 @@ int job_validate_file(JOB *job)
 	return 0;
 }
 
-JOB *job_init(JOB *job, wget_iri *iri, bool http_fallback)
+JOB *job_init(JOB *job, blacklist_entry *blacklistp, bool http_fallback)
 {
 	static unsigned long long jobid;
 
@@ -286,7 +285,8 @@ JOB *job_init(JOB *job, wget_iri *iri, bool http_fallback)
 	else
 		memset(job, 0, sizeof(JOB));
 
-	job->iri = iri;
+	job->blacklist_entry = blacklistp;
+	job->iri = blacklistp->iri; // convenience assignment
 	job->http_fallback = http_fallback;
 	job->id = ++jobid;
 
